/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>

#include <sys/types.h> /* regex_t */
#include <regex.h>

#include "intl.h"

#include "x_alloc.h"
#include "logging.h"
#include "preferences.h"

#define XEP_REGEXP_ERR_BUFSIZE 100

static char *
Preferences_subject_new(const char *subject, 
			const char *regexp, 
			const char *prefix);

static XePPreferences *global_preferences;

XePPreferences *
a_Preferences_create_from_tree(cntree *ntree)
{
  XePPreferences *preferences;

  preferences = x_alloc(1, XePPreferences);
  if (preferences)
    {
      preferences->prefs[XEP_PREFS_DEFAULT_CHARSET] = "ISO-8859-15";

      preferences->prefs[XEP_PREFS_EMPTY_SUBJECT] = _("(no subject)");

      preferences->prefs[XEP_PREFS_REPLY_REGEXP] = "re *: *|ant *: *";
      preferences->prefs[XEP_PREFS_REPLY_PREFIX] = "Re: ";

      preferences->prefs[XEP_PREFS_FORWARD_REGEXP] = "fw *: *|ref *: *";
      preferences->prefs[XEP_PREFS_FORWARD_PREFIX] = "Fw: ";
    }
  global_preferences = preferences;

  return global_preferences;
}

void
a_Preferences_free(XePPreferences *preferences)
{
  if (!preferences)
    return;

  free(preferences);    
  global_preferences = NULL;
}

void *
a_Preferences_get(XePPreferencesToken token)
{
  return global_preferences->prefs[token];
}

char *
a_Preferences_reply_subject_new(const char *subject)
{
  const char *regexp;
  const char *prefix;
 
  regexp = a_Preferences_get(XEP_PREFS_REPLY_REGEXP);
  prefix = a_Preferences_get(XEP_PREFS_REPLY_PREFIX);
  return Preferences_subject_new(subject, regexp, prefix);
}

char *
a_Preferences_forward_subject_new(const char *subject)
{
  const char *regexp;
  const char *prefix;
 
  regexp = a_Preferences_get(XEP_PREFS_FORWARD_REGEXP);
  prefix = a_Preferences_get(XEP_PREFS_FORWARD_PREFIX);
  return Preferences_subject_new(subject, regexp, prefix);
}

static char *
Preferences_subject_new(const char *subject, 
			const char *regexp, 
			const char *prefix)
{
  char *new_subject;
  regex_t preg;
  int r;
  const char *p = subject;

  if (!p || !*p)
    {
      p = a_Preferences_get(XEP_PREFS_EMPTY_SUBJECT);
      p = p ? p : "";
    }

  r = regcomp(&preg, regexp, REG_EXTENDED | REG_ICASE);
  if (r == 0)
    {
      do
	{
	  regmatch_t pmatch[1];
	  r = regexec(&preg, p, 1, pmatch, 0);
	  if (r == 0)
	    {
	      /* Accepts matches only if they are prefixes */
	      if (pmatch->rm_so == 0)
		p += pmatch->rm_eo;
	      else
		r = REG_NOMATCH;
	    }
	} while (r == 0);
    }
  else
    {
      char errbuf[XEP_REGEXP_ERR_BUFSIZE];
      regerror(r, &preg, errbuf, sizeof(errbuf));
      A_LOGGING_ERROR(_("Regular expression error for "
			"new subject creation: %s"),
		      errbuf);
    }

  regfree(&preg);

  prefix = (prefix ? prefix : "");
  new_subject = x_alloc(strlen(prefix) + strlen(p) + 1, char);
  strcpy(new_subject, prefix);
  strcat(new_subject, p);
  
  return new_subject;
}
