/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
 
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h> /* pipe, dup2, write, close, fork, setpgid, getppid,
		       sysconf, execv, mmap */
#include <sys/types.h> /* open, waitpid */
#include <sys/stat.h>  /* open */
#include <fcntl.h>     /* open */
#include <sys/mman.h>  /* mmap */

#include "intl.h"

#include "logging.h"
#include "x_alloc.h"
#include "defines.h"
#include "xetpan.h"
#include "remote.h"
#include "ui.h"
#include "ui-about.h"
#include "utils.h"

#include "folder-types.h"
#include "message.h"

/*! Define the size of the XePUIMain hash table
 *  Considering that rarely more than 8 UI will be open at the same
 *  time
 */
#define XEP_UIMAIN_HASH_SIZE 8

#define XEP_MESSAGEVIEWS_HASH_SIZE 16

typedef struct XePUIFunctions_s XePUIFunctions;
 
struct XePUIFunctions_s 
{

  /* Main loop functions */

  void (*init)          (XetPan *xetpan, int* argc, char ***argv);  
  int  (*init_check)    (XetPan *xetpan, int *argc, char ***argv);
  void (*main)          (XetPan *xetpan);
  void (*main_quit)     (XetPan *xetpan);
  int  (*events_pending)(XetPan *xetpan);
  void (*main_iteration)(XetPan *xetpan);

  /* First launch */

  void (*first_launch)(XetPan *xetpan);

  /* Main view */

  XePUIMain* (*mainview_create)(XetPan *xetpan, char *terminal);
  void (*mainview_close)       (XePUIMain *uimain);

  /* Folders List View */
  void (*foldersview_list_updated)(XePUIFoldersView *fview);
  void (*foldersview_list_refresh)(XePUIFoldersView *fview);

  /* Folder */

  void (*folder_updated)(XePUIMain *uimain, XePFolder *folder);

  /*! Redisplay the folder, mainly when messages flags are changed */
  void (*folder_refresh)(XePUIMain *uimain, XePFolder *folder);

  /* Message */

  void (*message_updated)(XePUIMain *uimain, XePMessage *message);
  /* Message composition */

  void (*compose_pending_requests)           (XetPan *xetpan);
  XePUIComposeWindow* (*composewindow_create)(XetPan *xetpan,
					      char *terminal);
  void (*composewindow_set_message)          (XePUIComposeWindow *uicompose, 
					      XePMessage *message);
  void (*composewindow_set_body_text)        (XePUIComposeWindow *uicompose,
					      const char *text,
					      size_t text_size,
					      const char *text_charset);
  void (*composewindow_add_header)           (XePUIComposeWindow *uicompose,
					      XePMessageHeaderType header_type,
					      const char *header_name,
					      const char *header_value);
  void (*composewindow_close)                (XePUIComposeWindow *uicompose);

  /* About dialog */

  void (*about)(XePUIMain *uimain);

  /* Error notification */

  void (*notify_error)      (XePUIMain *uimain);
  void (*notify_error_reset)(XePUIMain *uimain);

  /* Message view */

  void (*messageview_mime_view_show)                 (XePUIMessageView *mview, 
						      XePMessage *message);
  void (*messageview_mime_view_block)                (XePUIMessageView *mview,
					              XePMessage *message);
  void (*messageview_mime_view_unblock)              (XePUIMessageView *mview, 
						      XePMessage *message);
  void (*messageview_mime_show_selected)             (XePUIMessageView *mview,
						      XePMessage *message);
  void (*messageview_finalize)                       (XePUIMessageView *mview,
						      XePMessage *message);
  XePUIMessageView * (*messageview_create_standalone)(XetPan *xetpan, 
						      char *terminal);
  void (*messageview_close)                          (XePUIMessageView *mview);

  /* IO asynchroneous functions */

  int (*create_io_watch) (XePUI *ui, XePUIIOWatcher *watcher);
  void (*delete_io_watch)(XePUI *ui, XePUIIOWatcher *watcher);
};

static XePUIFunctions UI[XETPAN_UI_NB];

static cinthash *global_mainviews;
static cinthash *global_foldersviews;
static cinthash *global_messageviews;

XePUI *global_ui;

static XePUIMain *
UI_mainview_get_from_uid(XePUID uid);

static void
UI_mainview_close_all_func(void *auimain, void *anull);

static void
UI_folder_expunged(void *anull, void *afolder);

static void
UI_folder_processed(void *anull, void *afolder_processing);

static void
UI_foldersview_list_updated_func(void *auimain, void *anull);

static void
UI_foldersview_list_refresh_func(void *auimain, void *anull);

static void
UI_folder_updated_func(void *auimain, void *afolder);

static void
UI_message_updated_func(void *auimain, void *amessage);

static void
UI_folder_refresh_func(void *auimain, void *afolder);

static void
UI_notify_error_func(void *auimain, void *anull);

static void
UI_notify_error_reset_func(void *auimain, void *anull);

static void
UI_folder_messages_change_flags_cb(void *afolder, void *amessages);

static void
UI_messageview_finalize_cb(void *amview_uid, void *amessage);

static XePUIMessageView *
UI_messageview_get_from_uid(XePUID uid);

static void
UI_messageview_close_all_func(void *amview, void *auitype);

void
a_UI_messageview_fetch_show_async(XePUIMessageView *mview, 
				  XePMessage *message);
static void
UI_messageview_flags_change_async_cb(void *amview_uid, void *amessage);

static void
UI_messageview_mime_view_show_cb(void *axetpan, void *amessage);

static void
UI_messageview_mime_view_show_func(void *amview, void *amessage);

static void
UI_messageview_mime_view_block_cb(void *axetpan, void *amessage);

static void
UI_messageview_mime_view_block_func(void *amview, void *amessage);

static void
UI_messageview_mime_view_unblock_cb(void *axetpan, void *amessage);

static void
UI_messageview_mime_view_unblock_func(void *amview, void *amessage);

static void
UI_messageview_mime_show_selected_cb(void *amview, void *amessage);

static unsigned int
UI_get_free_watch_id(XePUI *ui);

/****************************************************************************/
/*                                                                          */
/* UI basic                                                                 */
/*                                                                          */
/****************************************************************************/

XePUI *
a_UI_new(void)
{
  XePUI *ui;
#ifdef USE_GTKUI
  XePUIFunctions gtkuifunctions =
    {
      a_GtkUI_init,
      a_GtkUI_init_check,
      a_GtkUI_main,
      a_GtkUI_main_quit,
      a_GtkUI_events_pending,
      a_GtkUI_main_iteration,
      a_GtkUI_first_launch,
      a_GtkUI_mainview_create,
      a_GtkUI_mainview_close,
      a_GtkUI_foldersview_list_updated,
      a_GtkUI_foldersview_list_refresh,
      a_GtkUI_folder_updated,
      a_GtkUI_folder_refresh,
      a_GtkUI_message_updated,
      a_GtkUI_compose_pending_requests,
      a_GtkUI_composewindow_create,
      a_GtkUI_composewindow_set_message,
      a_GtkUI_composewindow_set_body_text,
      a_GtkUI_composewindow_add_header,
      a_GtkUI_composewindow_close,
      a_GtkUI_about,
      a_GtkUI_notify_error,
      a_GtkUI_notify_error_reset,
      a_GtkUI_messageview_mime_view_show,
      a_GtkUI_messageview_mime_view_block,
      a_GtkUI_messageview_mime_view_unblock,
      a_GtkUI_messageview_mime_show_selected,
      a_GtkUI_messageview_finalize,
      a_GtkUI_messageview_create_standalone,
      a_GtkUI_messageview_close,
      a_GtkUI_create_io_watch,
      a_GtkUI_delete_io_watch
    };
#endif
#ifdef USE_TTYUI
  XePUIFunctions ttyuifunctions = 
    {
      a_TtyUI_init,
      a_TtyUI_init_check,
      a_TtyUI_main,
      a_TtyUI_main_quit,
      a_TtyUI_events_pending,
      a_TtyUI_main_iteration,
      a_TtyUI_first_launch,
      a_TtyUI_mainview_create,
      a_TtyUI_mainview_close,
      a_TtyUI_foldersview_list_updated,
      a_TtyUI_foldersview_list_refresh,
      a_TtyUI_folder_updated,
      a_TtyUI_folder_refresh,
      a_TtyUI_message_updated,
      a_TtyUI_compose_pending_requests,
      a_TtyUI_composewindow_create,
      a_TtyUI_composewindow_set_message,
      a_TtyUI_composewindow_set_body_text,
      a_TtyUI_composewindow_add_header,
      a_TtyUI_composewindow_close,
      a_TtyUI_about,
      a_TtyUI_notify_error,
      a_TtyUI_notify_error_reset,
      a_TtyUI_messageview_mime_view_show,
      a_TtyUI_messageview_mime_view_block,
      a_TtyUI_messageview_mime_view_unblock,
      a_TtyUI_messageview_mime_show_selected,
      a_TtyUI_messageview_finalize,
      a_TtyUI_messageview_create_standalone,
      a_TtyUI_messageview_close,
      a_TtyUI_create_io_watch,
      a_TtyUI_delete_io_watch
    };
#endif
#ifdef USE_GTKUI
  UI[XETPAN_UI_GTK] = gtkuifunctions;
#endif
#ifdef USE_TTYUI
  UI[XETPAN_UI_TTY] = ttyuifunctions;
#endif

  ui = x_alloc(1, XePUI);
  if (!ui)
    {
      A_LOGGING_SYSTEM_ERROR("calloc: %s", strerror(errno));
      return NULL;
    }
  ui->type = XETPAN_UI_UNSPECIFIED;
  ui->terminal = NULL;
  
  ui->mainviews = cinthash_new(XEP_UIMAIN_HASH_SIZE);
  if (!ui->mainviews)
    {
      A_LOGGING_SYSTEM_ERROR("cinthash_new()");
      goto free_ui;
    }
  global_mainviews = ui->mainviews;

  ui->foldersviews = cinthash_new(XEP_UIMAIN_HASH_SIZE);
  if (!ui->foldersviews)
    {
      A_LOGGING_SYSTEM_ERROR("cinthash_new()");
      goto free_ui;
    }
  global_foldersviews = ui->foldersviews;

  ui->composewindows = clist_new();
  if (!ui->composewindows)
    {
      A_LOGGING_SYSTEM_ERROR("clist_new()");
      goto free_mainviews;
    }

  ui->watched_fd = carray_new(10);
  if (!ui->watched_fd)
    {
      A_LOGGING_SYSTEM_ERROR("carray_new(): %s", strerror(errno));
      goto free_composewindows;
    }

  ui->messageviews = cinthash_new(XEP_MESSAGEVIEWS_HASH_SIZE);
  if (!ui->messageviews)
    {
      A_LOGGING_SYSTEM_ERROR("cinthash_new");
      goto free_watched_fd;
    }
  global_messageviews = ui->messageviews;

  ui->max_wid  = 0;
  ui->ideditor_lock = 0;
  ui->tty = NULL;
  ui->gtk = NULL;
  ui->switch_state = 1; /* No UI yet */

  global_ui = ui;

  return ui;

 free_watched_fd:
  carray_free(ui->watched_fd);
 free_composewindows:
  clist_free(ui->composewindows);
 free_mainviews:
  cinthash_free(ui->mainviews);
 free_ui:
  x_free(ui);
  global_ui = NULL;
  return NULL;
}

void
a_UI_free(XePUI *ui)
{
  clist_free(ui->composewindows);
  cinthash_free(ui->mainviews);
  carray_free(ui->watched_fd);
  x_free(ui->terminal);
  x_free(ui);
  global_ui = NULL;
}

int
a_UI_id_editor_is_locked(void)
{
  return (global_ui->ideditor_lock != 0);
}

int
a_UI_id_editor_lock(void)
{
  if (global_ui->ideditor_lock == 0)
    {
      global_ui->ideditor_lock = 1;
      return 0;
    }
  else
    {
      return -1;
    }
}

void
a_UI_id_editor_unlock(void)
{
  global_ui->ideditor_lock = 0;
}

int
a_UI_init(XetPan *xetpan, int *argc, char ***argv)
{
  int ret = 0;

  A_LOGGING_TRACE;

  if (xetpan->ui->type >= 0 && xetpan->ui->type < XETPAN_UI_NB)
    {
      UI[xetpan->ui->type].init(xetpan, argc, argv);
    }
  else
  {
    int uitype;
    for (uitype = 0; uitype < XETPAN_UI_NB; uitype++)
      {
        if (UI[uitype].init_check(xetpan, argc, argv) == 0)
          {
            xetpan->ui->type = uitype;
            break;
          }
      }
    if (uitype == XETPAN_UI_NB)
      {
        A_LOGGING_ERROR("No usable User Interface could be found");

        xetpan->ui->type = XETPAN_UI_QUIT;
        ret = -1;
      }
  }


  if (ret == 0 && xetpan->ui->switch_state)
    {
      carray *watched_fds =  xetpan->ui->watched_fd;
      XePUIIOWatcher *watcher;
      int i;
      xetpan->ui->switch_state = 0;
      for (i = 0; i  < carray_count(watched_fds); i++)
	{
	  watcher = (XePUIIOWatcher *) carray_get(watched_fds, i);
	  UI[xetpan->ui->type].create_io_watch(xetpan->ui, watcher);
	}
    }   
      
      return ret;
}

void
a_UI_main(XetPan *xetpan)
{
    int argc = 0;

    A_LOGGING_TRACE;

    while (xetpan->ui->type != XETPAN_UI_QUIT)
      {
        A_LOGGING_DEBUG("inner loop, ui type = %d", xetpan->ui->type);
        if (xetpan->ui->switch_state)
          {
            a_UI_init(xetpan, &argc, NULL);
            a_UI_mainview_create(xetpan, xetpan->ui->terminal, 
                                 xetpan->ui->type);
          }
        else if (xetpan->ui->type >= 0 && xetpan->ui->type < XETPAN_UI_NB) 
          {
            UI[xetpan->ui->type].main(xetpan);
          }
        else
          {
	    A_LOGGING_SYSTEM_ERROR("UI type %d unknown. Quitting.", 
				   xetpan->ui->type);
	    xetpan->ui->type = XETPAN_UI_QUIT;
          }
      }
}

void
a_UI_main_quit(XetPan *xetpan)
{
  a_UI_main_switch(xetpan, NULL, XETPAN_UI_QUIT);
}

int
a_UI_main_switch(XetPan *xetpan, const char *terminal, int uitype)
{
  carray *watched_fds;
  XePUIIOWatcher *watcher;
  XePUI *ui = xetpan->ui;
  int ret_val = 0;
  int n_watches;

  /* TODO: see if we can avoid returning a value */

  if (ui->type == uitype)
    return ret_val;

  watched_fds = ui->watched_fd;
  
  n_watches =  carray_count(watched_fds);

  A_LOGGING_INFO("Switching to a User Interface");
  A_LOGGING_DEBUG("switching to UI type %d", uitype);

  UI[ui->type].main_quit(xetpan);

  if (uitype == XETPAN_UI_QUIT)
    {
      A_LOGGING_DEBUG("quitting and deleting IO watchers");
      
      while (carray_count(watched_fds))
        {
          watcher = (XePUIIOWatcher *) carray_delete(watched_fds, 0);
          x_free(watcher);
        }
    }

  if (ret_val == 0 && uitype != XETPAN_UI_QUIT)
    {
      ui->type = uitype;
      ui->switch_state = 1;
      x_free(ui->terminal);
      if (terminal)
	ui->terminal = strdup(terminal);
      else
	ui->terminal = NULL;
    }

  if (uitype == XETPAN_UI_QUIT)
    ui->type = XETPAN_UI_QUIT;

  return 0;
}

int
a_UI_events_pending(XetPan *xetpan)
{
  A_LOGGING_TRACE;
  return UI[xetpan->ui->type].events_pending(xetpan);
}

void
a_UI_main_iteration(XetPan *xetpan)
{
  A_LOGGING_TRACE;
  UI[xetpan->ui->type].main_iteration(xetpan);
}

void
a_UI_wait_until(XetPan *xetpan, int *semaphore, int value)
{
  A_LOGGING_TRACE;
  
  while ((*semaphore != value))
    {
      printf("sem = %d val = %d\n", *semaphore, value);
      a_UI_main_iteration(xetpan);
    }
}

/****************************************************************************/
/*                                                                          */
/* UI: windows (GUI) and screens (TTY)                                      */
/*                                                                          */
/****************************************************************************/

void
a_UI_mainview_init(XePUIMain *uimain)
{
  static XePUID uid = 0;

  uimain->uid         = uid++;
  uimain->toplevel    = NULL;
  uimain->foldersview = NULL;
  uimain->summaryview = NULL;
  uimain->messageview = NULL;
  uimain->menubar     = NULL;
  uimain->toolbar     = NULL;
  uimain->statusbar   = NULL;
  uimain->about       = NULL;
  uimain->ideditor    = NULL;
  uimain->xetpan      = NULL;
  uimain->remote      = NULL;
  uimain->python      = NULL;
  uimain->logview     = NULL;
  uimain->kb_watch_id = 0;

  cinthash_add(global_mainviews, uimain->uid, uimain);
}

static XePUIMain *
UI_mainview_get_from_uid(XePUID uid)
{
  return (XePUIMain *) cinthash_find(global_mainviews, uid);
}

void
a_UI_foldersview_init(XePUIFoldersView *fview)
{
  static XePUID uid = 0;

  fview->uid = uid++;

  cinthash_add(global_foldersviews, fview->uid, fview);
}

void
a_UI_foldersview_unregister(XePUIFoldersView *fview)
{
  cinthash_remove(global_foldersviews, fview->uid);
}

XePUIComposeWindow *
a_UI_composewindow_new(void)
{
  XePUIComposeWindow *uicompose;

  uicompose = x_alloc(1, XePUIComposeWindow);
  if (uicompose == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
      return NULL;
    }
  uicompose->any.type         = XETPAN_UI_UNSPECIFIED;
  uicompose->any.xetpan       = NULL;
  uicompose->any.uimain       = NULL;
  uicompose->any.mta          = NULL;
  uicompose->any.messageview  = NULL;
  uicompose->any.sent_folder  = NULL;
  return uicompose;
}

XePUIMessageView *
a_UI_messageview_new(void)
{
  XePUIMessageView *mview;
  static XePUID uid = 0;

  mview = x_alloc(1, XePUIMessageView);
  if (mview == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
      return NULL;
    }

  mview->any.type     = XETPAN_UI_UNSPECIFIED;
  mview->any.uid      = uid++;
  mview->any.editable = 0;
  mview->any.xetpan   = NULL;
  mview->any.uimain   = NULL;
  mview->any.folder   = NULL;
  mview->any.message  = NULL;
  mview->any.muid     = 0;
  mview->any.inlined_list    = NULL;
  mview->any.body_text       = NULL;
  mview->any.default_charset = NULL;

  cinthash_add(global_messageviews, mview->any.uid, mview);

  return mview;
}

static XePUIMessageView *
UI_messageview_get_from_uid(XePUID uid)
{
  return (XePUIMessageView *) cinthash_find(global_messageviews, uid);
}

void
a_UI_first_launch(XetPan *xetpan)
{

  A_LOGGING_TRACE;
  UI[xetpan->ui->type].first_launch(xetpan);
}

/* Compose a list of messages, then if there is no main view, offer to
 * create one, else exit.
 * DEPRECATED
 */
void
a_UI_compose_pending_requests(XetPan *xetpan)
{
  A_LOGGING_TRACE;
  UI[xetpan->ui->type].compose_pending_requests(xetpan);
}

/* Create main view */
XePUIMain *
a_UI_mainview_create(XetPan *xetpan, char *terminal, int type)
{
  XePUIMain *uimain;

  A_LOGGING_TRACE;

  uimain = UI[type].mainview_create(xetpan, terminal);

  return uimain;
}

int
a_UI_mainview_close(XePUIMain *uimain)
{
  cinthash *mainviews;
  XetPan *xetpan;

  A_LOGGING_TRACE;

  xetpan = (XetPan *) uimain->xetpan;

  mainviews = xetpan->ui->mainviews;

  cinthash_remove(mainviews, uimain->uid);

  UI[uimain->type].mainview_close(uimain);

  return a_XetPan_should_end(xetpan);
}

int
a_UI_mainview_close_all(XetPan *xetpan, int uitype)
{
  A_LOGGING_TRACE;

  cinthash_foreach_data(xetpan->ui->mainviews, 
			UI_mainview_close_all_func, NULL);

  cinthash_free(xetpan->ui->mainviews);
  xetpan->ui->mainviews = cinthash_new(XEP_UIMAIN_HASH_SIZE);
  global_mainviews = xetpan->ui->mainviews;

  return a_XetPan_should_end(xetpan);
}

static void
UI_mainview_close_all_func(void *auimain, void *anull)
{
  XePUIMain *uimain = (XePUIMain *) auimain;

  UI[uimain->type].mainview_close(uimain);
}

void
a_UI_expunge_folder(const char *ref)
{
  XePFolder *folder;
  XePManagerMsg *msg;

  A_LOGGING_DEBUG("Expunging folder '%s'", ref);

  folder = a_Folder_get_from_ref(ref);
  if (folder)
    {
      switch (folder->type)
	{
	case XEP_FOLDER_TYPE_REAL:
	  msg = a_Manager_msg_new();
	  a_Manager_msg_set_type(msg, XEP_MANAGER_EXPUNGE);
	  a_Manager_msg_set_source_target(msg, NULL, &folder->thread);
	  a_Manager_msg_set_args(msg, folder, folder);
	  a_Manager_msg_add_callback(msg, UI_folder_expunged, NULL);
	  
	  a_Manager_msg_submit(msg);
	  break;
	case XEP_FOLDER_TYPE_MAILBOX:
	  break;
	case XEP_FOLDER_TYPE_VIRTUAL:
	  break;
	}
    }
  else
    {
      A_LOGGING_ERROR("Unknown folder '%s'", ref);
    }

}

static void
UI_folder_expunged(void *anull, void *afolder)
{
  XePFolder *folder = (XePFolder *) afolder;
  a_UI_update_folder(folder->id);
}

void
a_UI_update_folder(const char *ref)
{
  XePFolder *folder;
  XePManagerMsg *msg;
  A_LOGGING_DEBUG("Updating folder '%s'", ref);

  if (ref == NULL)
    {
      a_UI_folder_updated(NULL, NULL);
      return;
    }

  folder = a_Folder_get_from_ref(ref);
  if (folder)
    {
      switch (folder->type)
	{
	case XEP_FOLDER_TYPE_REAL:
	  msg = a_Manager_msg_new();
	  a_Manager_msg_set_type(msg, XEP_MANAGER_UPDATE);
	  a_Manager_msg_set_source_target(msg, NULL, &folder->thread);
	  a_Manager_msg_set_args(msg, folder, folder);
	  a_Manager_msg_add_callback(msg, a_UI_folder_updated, NULL);
	  
	  a_Manager_msg_submit(msg);
	  break;
	case XEP_FOLDER_TYPE_MAILBOX:
	  break;
	case XEP_FOLDER_TYPE_VIRTUAL:
	  break;
	}
    }
  else
    {
      A_LOGGING_ERROR("Unknown folder '%s'", ref);
    }
}

void
a_UI_refresh_folder(const char *ref)
{
  XePFolder *folder;
  A_LOGGING_TRACE;

  folder = a_Folder_get_from_ref(ref);

  if (folder)
    cinthash_foreach_data(global_mainviews, UI_folder_refresh_func, folder);
}

static void
UI_folder_refresh_func(void *auimain, void *afolder)
{
  XePUIMain *uimain = (XePUIMain *) auimain;
  XePFolder *folder = (XePFolder *) afolder;

  UI[uimain->type].folder_refresh(uimain, folder);
}

void
a_UI_folder_process(XePFolder *folder, 
		    carray *messages, 
		    clist *processing_rules)
{
  XePFolderProcessing *folder_processing;
  XePManagerMsg *msg;
  
  folder_processing = a_Folder_processing_new(folder, messages, 
					      processing_rules);
  msg = a_Manager_msg_new();
  a_Manager_msg_set_type(msg, XEP_MANAGER_PROCESSING);
  a_Manager_msg_set_source_target(msg, NULL, &folder->thread);
  a_Manager_msg_set_args(msg, folder_processing, folder_processing);
  a_Manager_msg_add_callback(msg, UI_folder_processed, folder);
  a_Manager_msg_submit(msg);
}

static void
UI_folder_processed(void *afolder, void *afolder_processing)
{
  XePFolder *folder = (XePFolder *) afolder;
  XePFolderProcessing *folder_processing =
    (XePFolderProcessing *) afolder_processing;

  a_Folder_processing_free(folder_processing);

  /* Processing may have affected other folders */
  a_UI_foldersview_list_refresh();

  /* Processing may have affected the displayed folder */
  /* FIXME: We need to call a_UI_folder_updated(), in fact.
   * This will be done when folder update will retain
   * the current view state
   */
  a_UI_refresh_folder(a_Folder_get_id(folder));
}

void
a_UI_foldersview_list_updated(void)
{
  A_LOGGING_TRACE;

  cinthash_foreach_data(global_foldersviews, UI_foldersview_list_updated_func,
			NULL);
}

static void
UI_foldersview_list_updated_func(void *afview, void *anull)
{
  XePUIFoldersView *fview = (XePUIFoldersView *) afview;

  UI[fview->type].foldersview_list_updated(fview);
}

void
a_UI_foldersview_list_refresh(void)
{
  A_LOGGING_TRACE;

  cinthash_foreach_data(global_foldersviews, UI_foldersview_list_refresh_func,
			NULL);
}

static void
UI_foldersview_list_refresh_func(void *afview, void *anull)
{
  XePUIFoldersView *fview = (XePUIFoldersView *) afview;

  UI[fview->type].foldersview_list_refresh(fview);
}

void
a_UI_folder_updated(void *anull, void *afolder)
{
  A_LOGGING_TRACE;

  a_UI_foldersview_list_refresh();
  cinthash_foreach_data(global_mainviews, UI_folder_updated_func, afolder);

}

static void
UI_folder_updated_func(void *auimain, void *afolder)
{
  XePUIMain *uimain = (XePUIMain *) auimain;
  XePFolder *folder = (XePFolder *) afolder;

  UI[uimain->type].folder_updated(uimain, folder);
}

void
a_UI_message_updated(XePMessage* message)
{
  A_LOGGING_TRACE;

  cinthash_foreach_data(global_mainviews, UI_message_updated_func, message);
}

static void
UI_message_updated_func(void *auimain, void *amessage)
{
  XePUIMain *uimain = (XePUIMain *) auimain;
  XePMessage *message = (XePMessage *) amessage;

  UI[uimain->type].message_updated(uimain, message);
}

/* Create a compose window */
XePUIComposeWindow *
a_UI_composewindow_create(XetPan *xetpan, char *terminal, int type)
{
  XePUIComposeWindow *uicompose;

  A_LOGGING_TRACE;

  uicompose = UI[type].composewindow_create(xetpan, terminal);

  /* Append this compose window to the list of compose windows */

  clist_append(xetpan->ui->composewindows, uicompose);

  return uicompose;
}

XePUIComposeWindow *
a_UI_composewindow_compose(XetPan *xetpan, 
			   char *terminal, 
			   int uitype,
			   const char *mailto)
{
  XePUIComposeWindow *uicompose;
  XePMessage *message;
  char *initial_text;

  uicompose = a_UI_composewindow_create(xetpan, terminal, uitype);
  
  message = a_Message_new_to_compose();

  a_UI_composewindow_set_message(uicompose, message);

  if (mailto)
    {
      a_UI_composewindow_set_mailto(uicompose, mailto);
    }
  else
    {
      a_UI_composewindow_add_header(uicompose, XEP_MAIL_HEADER_TO, NULL, 
				    NULL);
    }

#if 0
  initial_text = a_Templates_();
  a_UI_composewindow_set_body_text(uicompose, initial_text, GTK_DEST_CHARSET);
#endif

  return uicompose;
}
			
XePUIComposeWindow *
a_UI_composewindow_reply(XetPan *xetpan, 
			 char *terminal, 
			 int uitype,
			 XePMessage *original_message,
			 const char *reply_text,
			 const char *reply_charset)
{
  XePUIComposeWindow *uicompose;
  XePMessage *reply_message;

  uicompose = a_UI_composewindow_create(xetpan, terminal, uitype);
  
  /* TODO: This should be done asynchronously */
  /* TODO: If reply_text is given, use it for the reply,
   * else fetch the default part of the message
   */

  reply_message = a_Message_new_to_reply(original_message);
  
  a_UI_composewindow_set_message(uicompose, reply_message);

  if (reply_text)
    {
      a_UI_composewindow_set_body_text(uicompose,
				       reply_text, 
				       strlen(reply_text),
				       reply_charset);
    }
  else
    {
      /* TODO: the following should be done in a callback function
       * called when fetching 
       */
      a_Message_mime_find_default(original_message);
      a_Message_fetch_part(original_message);
      if (original_message->data)
	{
	  a_UI_composewindow_set_body_text(uicompose,
					   original_message->data, 
					   original_message->len,
					   original_message->source_charset);
	  x_free(original_message->data);
	  original_message->data = NULL;
	}
    }

  return uicompose;
}

void
a_UI_composewindow_set_mailto(XePUIComposeWindow *uicompose,
			      const char *mailto)
{
  clist *headers;
  clistiter *iter;
  XePMessageHeaderType type;

  headers = a_Utils_parse_mailto(mailto);

  for (iter = clist_begin(headers); iter; iter = iter->next)
    {
      XePUtilsPair *header = (XePUtilsPair *) iter->data;
      type = a_Message_header_get_type_from_name(header->name);
      UI[uicompose->any.type].composewindow_add_header(uicompose, type,
						       header->name,
						       header->value);
      x_free(header->name);
      x_free(header->value);
      x_free(header);
    }
  clist_free(headers);
}

void
a_UI_composewindow_set_body_text(XePUIComposeWindow *uicompose,
				 const char *text,
				 size_t text_size,
				 const char *text_charset)
{
  UI[uicompose->any.type].composewindow_set_body_text(uicompose,
						      text,
						      text_size,
						      text_charset);
}

XePUIComposeWindow *
a_UI_composewindow_forward(XetPan *xetpan, 
			   char *terminal, 
			   int uitype,
			   XePMessage *message)
{
  XePUIComposeWindow *uicompose;
  XePMessage *forward_message;
  uicompose = a_UI_composewindow_create(xetpan, terminal, uitype);

  /* This should be done asynchronously */
  forward_message = a_Message_new_to_forward(message);

  a_UI_composewindow_set_message(uicompose, forward_message);
  return uicompose;
}

XePUIComposeWindow *
a_UI_composewindow_redirect(XetPan *xetpan, 
			    char *terminal, 
			    int uitype,
			    XePMessage *message)
{
  XePUIComposeWindow *uicompose;
  XePMessage *redirect_message;

  uicompose = a_UI_composewindow_create(xetpan, terminal, uitype);
  
  /* This should be done asynchronously */
  redirect_message = a_Message_new_to_redirect(message);

  a_UI_composewindow_set_message(uicompose, redirect_message);
  return uicompose;
}

void
a_UI_composewindow_set_message(XePUIComposeWindow *uicompose,
			       XePMessage *message)
{
  uicompose->any.messageview->any.muid = message->muid;
  uicompose->any.messageview->any.message = message;
  UI[uicompose->any.type].composewindow_set_message(uicompose, message);
}

void
a_UI_composewindow_add_header(XePUIComposeWindow *uicompose, 
			      XePMessageHeaderType header_type,
			      const char *header_name,
			      const char *header_value)
{
  UI[uicompose->any.type].composewindow_add_header(uicompose,
						   header_type,
						   header_name,
						   header_value);
}

int
a_UI_composewindow_close(XePUIComposeWindow *uicompose)
{
  clist *composewindows;
  clistiter *iter;
  clistiter *previous;
  XetPan *xetpan;

  A_LOGGING_TRACE;

  xetpan = (XetPan *) uicompose->any.xetpan;

  composewindows = xetpan->ui->composewindows;

  iter = clist_end(composewindows);

  while (iter)
    {
      previous = iter->previous;

      if (iter->data == uicompose)
	{
	  clist_delete(composewindows, iter);
          UI[uicompose->any.type].composewindow_close(uicompose);
          break;
	}

      iter = previous;
      }
  return a_XetPan_should_end(xetpan);
}

int
a_UI_composewindow_close_all(XetPan *xetpan, int uitype)
{
  clist *composewindows;
  clistiter *iter;
  clistiter *previous;
  XePUIComposeWindow *uicompose;

  A_LOGGING_TRACE;

  composewindows = xetpan->ui->composewindows;

  iter = clist_end(composewindows);
  while (iter)
    {
      uicompose = (XePUIComposeWindow *) iter->data;
      previous = iter->previous;

      if (uicompose->any.type == uitype)
	{
	  clist_delete(composewindows, iter);
          UI[uicompose->any.type].composewindow_close(uicompose);
	}
      iter = previous;
    }

 return a_XetPan_should_end(xetpan);
}

/*! Create a standalone message viewer */
XePUIMessageView *
a_UI_messageview_create(XetPan *xetpan, 
			char *terminal, 
			int  uitype)
{
	  return UI[uitype].messageview_create_standalone(xetpan, terminal);
}

/*! View a RFC822 file in a message viewer */
void
a_UI_messageview_view_file(XePUIMessageView *mview, char *filename)
{
  char *data;
  struct stat stat_buf;
  int fd;

  fd = open(filename, O_RDONLY);

  if (fd == -1)
    {
      A_LOGGING_ERROR(_("Error while opening file '%s': %s"),
		      filename, strerror(errno));
    }
  else
    {
      fstat(fd, &stat_buf);

      if (stat_buf.st_size)
	{
	  XePMessage *message;
	  char *mapdata;
	  mapdata = mmap(NULL, stat_buf.st_size, PROT_READ, MAP_SHARED, fd, 0);
	  data = x_alloc(stat_buf.st_size, char);
	  memcpy(data, mapdata, stat_buf.st_size);
	  message = a_Message_new_from_data(data, stat_buf.st_size);
	  /* Use the filename as the mailmessage::uid */
	  message->msg->msg_uid = strdup(filename);
	  a_UI_messageview_show_message(mview, message);
	  munmap(mapdata, stat_buf.st_size);
	}
      close(fd);
    }
}

int
a_UI_messageview_close(XePUIMessageView *mview)
{
  cinthash *messageviews;
  XetPan *xetpan;

  A_LOGGING_TRACE;

  xetpan = mview->any.xetpan;
  messageviews = xetpan->ui->messageviews;

  cinthash_remove(messageviews, mview->any.uid);

  UI[mview->any.type].messageview_close(mview);

  return a_XetPan_should_end(xetpan);
}

int
a_UI_messageview_close_all(XetPan *xetpan, int uitype)
{
  A_LOGGING_TRACE;

  cinthash_foreach_data(xetpan->ui->messageviews, 
			UI_messageview_close_all_func, (void *) uitype);

  /* FIXME: currently all message views  are closed independently of the uitype
   * so we can free the message views hashtable. 
   * TODO: selectively close message view depending on the uitype
   */
  cinthash_free(xetpan->ui->messageviews);
  xetpan->ui->messageviews = cinthash_new(XEP_MESSAGEVIEWS_HASH_SIZE);
  global_messageviews = xetpan->ui->messageviews;

  return a_XetPan_should_end(xetpan);
}

static void
UI_messageview_close_all_func(void *amview, void *auitype)
{
  XePUIMessageView *mview = (XePUIMessageView *) amview;
  
  /* TODO: Do selective close depending on UI type */
      UI[mview->any.type].messageview_close(mview);
}

int
a_UI_close_all(XetPan *xetpan, int uitype)
{
  a_UI_mainview_close_all(xetpan, uitype);
  return a_UI_composewindow_close_all(xetpan, uitype);
}

void
a_UI_notify_error(XetPan *xetpan)
{
  cinthash *mainviews;

  /* A_LOGGING_TRACE; Do not use logging because it will create an
   * endless loop
   */

  mainviews = xetpan->ui->mainviews;

  cinthash_foreach_data(mainviews, UI_notify_error_func, NULL);
}

static void
UI_notify_error_func(void *auimain, void *anull)
{
  XePUIMain *uimain = (XePUIMain *) auimain;

  UI[uimain->type].notify_error(uimain);
}

void
a_UI_notify_error_reset(XetPan *xetpan)
{
  cinthash *mainviews;

  /* A_LOGGING_TRACE; Do not use logging because it will create an
   * endless loop
   */

  mainviews = xetpan->ui->mainviews;

  cinthash_foreach_data(mainviews, UI_notify_error_reset_func, NULL);
}
 
static void
UI_notify_error_reset_func(void *auimain, void *anull)
{
  XePUIMain *uimain = (XePUIMain *) auimain;
  
  UI[uimain->type].notify_error_reset(uimain);
}

/***************************************************************************/
/*                                                                         */
/* Asynchronous message handling and display                               */
/*                                                                         */
/***************************************************************************/

void
a_UI_folder_messages_change_flags(XePFolder *folder, carray *messages)
{
  a_Folder_messages_change_flags_async(folder, messages, 
				       UI_folder_messages_change_flags_cb,
				       folder);
}

static void
UI_folder_messages_change_flags_cb(void *afolder, void *amessages)
{
  carray *messages = (carray *) amessages;
  XePFolder *folder = (XePFolder *) afolder;
  int i;
  int count = carray_count(messages);

  for (i = 0; i < count; i++)
    {
      a_Message_unref((XePMessage *) carray_get(messages, i));
    }
  carray_free(messages);
  a_UI_refresh_folder(folder->id);
  a_UI_foldersview_list_refresh();
}


void
a_UI_messageview_show_message(XePUIMessageView *mview,
			      XePMessage *message)
{
  mview->any.muid = message->muid;
  if (message->async)
    {
      a_UI_messageview_fetch_show_async(mview, message);
    }
  else
    {
      a_Message_mime_find_default(message);
      a_UI_messageview_mime_view_show(mview->any.xetpan, message);
      a_UI_messageview_mime_show_selected(mview, message);
      a_UI_messageview_finalize(mview, message);
    }
}

void
a_UI_messageview_fetch_show_async(XePUIMessageView *mview, 
				  XePMessage *message)
{
  XePManagerMsg *managermsg;
  XePFolder *folder = message->folder;

  A_LOGGING_TRACE;

  /* TODO: This message fetching flow is not optimal because
   * setting the message flags need to update the message mime tree
   * because setting flags using mailmessage_flush frees
   * temporary data like the mime tree.
   */

  managermsg = a_Manager_msg_new();
  if (managermsg == NULL)
    {
      return;
    }

  /* Cancel older operations and get bodystructure */
  a_Manager_msg_flush_old(managermsg);
  a_Manager_msg_set_type(managermsg, XEP_MANAGER_GET_BODYSTRUCTURE);
  a_Manager_msg_set_source_target(managermsg, NULL, &folder->thread);
  a_Manager_msg_set_args(managermsg, message, message);
  a_Manager_msg_add_callback(managermsg,
			     UI_messageview_mime_view_show_cb,
			     mview->any.xetpan);
  a_Manager_msg_submit(managermsg);

  managermsg = a_Manager_msg_new();
  if (managermsg == NULL)
    {
      return;
    }

  /* Find the default MIME part to show */
  a_Manager_msg_set_type(managermsg, XEP_MANAGER_MIME_FIND_DEFAULT);
  a_Manager_msg_set_source_target(managermsg, NULL, &folder->thread);
  a_Manager_msg_set_args(managermsg, message, message);
  a_Manager_msg_submit(managermsg);


  /* Fetch the default MIME part and show it when available 
   * then change the flags to SEEN/ NOT NEW
   * 
   */
  /* FIXME: XePMessage can possibly be freed while changing flags
   * TRYING: do the finalize in a_UI_messageview_flags_change_async
   *
   */
  message->flags_set = MAIL_FLAG_SEEN;
  message->flags_unset = MAIL_FLAG_NEW;

  managermsg = a_Manager_msg_new();
  if (managermsg == NULL)
    {
      return;
    }
  a_Manager_msg_set_type(managermsg, XEP_MANAGER_FETCH_PART);
  a_Manager_msg_set_source_target(managermsg, NULL, &folder->thread);
  a_Manager_msg_set_args(managermsg, message, message);
  a_Manager_msg_add_callback(managermsg,
			     UI_messageview_mime_show_selected_cb,
			     (void *) mview->any.uid);
  a_Manager_msg_add_callback(managermsg,
			     UI_messageview_flags_change_async_cb,
			     (void *) mview->any.uid);
  a_Manager_msg_submit(managermsg);
}

static void
UI_messageview_finalize_cb(void *amview_uid, void *amessage)
{
  XePUID mview_uid = (XePUID) amview_uid;
  XePMessage *message =(XePMessage *) amessage;
  XePUIMessageView *mview;

  mview = UI_messageview_get_from_uid(mview_uid);
  if (mview)
    {
      /* Notify that message changed (flags) */
      a_UI_message_updated(message);
      /* Notify that folder changed (message count) */
      a_UI_foldersview_list_refresh();
      if (message->folder)
	a_UI_refresh_folder(message->folder->id);
 
     a_UI_messageview_finalize(mview, message);
    }
  else
    {
      /* Take the responsability of freeing lost resources 
       * due to a main window that was closed
       */
      a_Message_unref(message);
    }
}

void
a_UI_messageview_finalize(XePUIMessageView *mview, XePMessage *message)
{
  UI[mview->any.type].messageview_finalize(mview, message);
}


static void
UI_messageview_flags_change_async_cb(void *amview_uid, void *amessage)
{
  XePUID mview_uid = (XePUID) amview_uid;
  XePMessage *message =(XePMessage *) amessage;
  XePUIMessageView *mview;

  mview = UI_messageview_get_from_uid(mview_uid);
  if (mview)
    {
      a_UI_messageview_flags_change_async(mview, message);
    }
  else
    {
      UI_messageview_finalize_cb(amview_uid, amessage);
    }
}

void
a_UI_messageview_flags_change_async(XePUIMessageView *mview, 
				    XePMessage *message)
{
  XePManagerMsg *managermsg;

  managermsg = a_Manager_msg_new();
  if (managermsg == NULL)
    {
      return;
    }

  a_Manager_msg_set_type(managermsg, XEP_MANAGER_FLAGS_CHANGE);
  a_Manager_msg_set_source_target(managermsg, NULL, &message->folder->thread);
  a_Manager_msg_set_args(managermsg, message, message);
  a_Manager_msg_add_callback(managermsg,
			     UI_messageview_finalize_cb,
			     (void *) mview->any.uid);

  a_Manager_msg_submit(managermsg);
}

static void
UI_messageview_mime_view_show_cb(void *axetpan, void *amessage)
{
  XetPan *xetpan = (XetPan *) axetpan;
  XePMessage *message =(XePMessage *) amessage;
  
  a_UI_messageview_mime_view_show(xetpan, message);
}

void
a_UI_messageview_mime_view_show(XetPan *xetpan, XePMessage *message)
{
  cinthash *messageviews;

  messageviews = xetpan->ui->messageviews;

  cinthash_foreach_data(messageviews, 
			UI_messageview_mime_view_show_func,
			message);
}

static void
UI_messageview_mime_view_show_func(void *amview, void *amessage)
{
  XePUIMessageView *mview = (XePUIMessageView *) amview;
  XePMessage *message =(XePMessage *) amessage;
  
  UI[mview->any.type].messageview_mime_view_show(mview, message);
}

#if 0
static void
UI_messageview_mime_view_block_cb(void *axetpan, void *amessage)
{
  XetPan *xetpan = (XetPan *) axetpan;
  XePMessage *message =(XePMessage *) amessage;
  
  a_UI_messageview_mime_view_block(xetpan, message);
}

void
a_UI_messageview_mime_view_block(XetPan *xetpan, XePMessage *message)
{
  cinthash *messageviews;

  messageviews = xetpan->ui->messageviews;

  cinthash_foreach_data(messageviews, 
			UI_messageview_mime_view_block_func, 
			message);
}
#endif

static void
UI_messageview_mime_view_block_func(void *amview, void *amessage)
{
  XePUIMessageView *mview = (XePUIMessageView *) amview;
  XePMessage *message =(XePMessage *) amessage;
  
  UI[mview->any.type].messageview_mime_view_block(mview, message);
}

static void
UI_messageview_mime_view_unblock_cb(void *axetpan, void *amessage)
{
  XetPan *xetpan = (XetPan *) axetpan;
  XePMessage *message =(XePMessage *) amessage;
  
  a_UI_messageview_mime_view_unblock(xetpan, message);
}


void
a_UI_messageview_mime_view_unblock(XetPan *xetpan, XePMessage *message)
{
  cinthash *messageviews;

  messageviews = xetpan->ui->messageviews;

  cinthash_foreach_data(messageviews,
			UI_messageview_mime_view_unblock_func,
			message);
}

static void
UI_messageview_mime_view_unblock_func(void *amview, void *amessage)
{
  XePUIMessageView *mview = (XePUIMessageView *) amview;
  XePMessage *message =(XePMessage *) amessage;
  
  UI[mview->any.type].messageview_mime_view_unblock(mview, message);
}

void
a_UI_messageview_mime_fetch_show_selected_async(XePUIMessageView *mview, 
						XePMessage *message)
{
  XePManagerMsg *managermsg;

  managermsg = a_Manager_msg_new();
  if (managermsg == NULL)
    {
      return;
    }

  a_Manager_msg_set_type(managermsg, XEP_MANAGER_FETCH_PART);
  a_Manager_msg_set_source_target(managermsg, NULL, &message->folder->thread);
  a_Manager_msg_set_args(managermsg, message, message);
  a_Manager_msg_add_callback(managermsg,
			     UI_messageview_mime_show_selected_cb,
			     (void *) mview->any.uid);

  a_Manager_msg_submit(managermsg);
}

static void
UI_messageview_mime_show_selected_cb(void *amview_uid, void *amessage)
{
  XePUID mview_uid = (XePUID) amview_uid;
  XePMessage *message =(XePMessage *) amessage;
  XePUIMessageView *mview;

  mview = UI_messageview_get_from_uid(mview_uid);
  if (mview)
    {
      a_UI_messageview_mime_show_selected(mview, message);
    }
}


void
a_UI_messageview_mime_show_selected(XePUIMessageView *mview,
				    XePMessage *message)
{
  UI[mview->any.type].messageview_mime_show_selected(mview, message);
}

/***************************************************************************/
/*                                                                         */
/* Watching I/O traffic                                                    */
/*                                                                         */
/***************************************************************************/

unsigned int
a_UI_add_io_watch(XetPan *xetpan,
		  int fd,
		  short int condition,
		  void (*callback)(int afd, short int acondition, void *adata),
		  void *data)
{
  XePUIIOWatcher *watcher;
  int ret;
  unsigned int watcher_count;

  watcher = x_alloc(1, XePUIIOWatcher);
  if (watcher == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("calloc: %s", strerror(errno));
      return 0;
    }

  /* Find an unused watch_id */
  watcher->watch_id  = UI_get_free_watch_id(xetpan->ui);

  watcher->fd        = fd;
  watcher->callback  = callback;
  watcher->data      = data;
  watcher->condition = condition;

  if (carray_add(xetpan->ui->watched_fd, watcher, &watcher_count) != 0)
    {
      A_LOGGING_SYSTEM_ERROR("carray_add failed");
      free(watcher);
      return 0;
    }
  
  /* This is done to have watch_id start at 1 not zero, as zero means error */


  A_LOGGING_DEBUG("added watch_id %d for fd %d",
		  watcher->watch_id, watcher->fd);
  if (xetpan->ui->switch_state == 0)
    {
      
      if ((ret = UI[xetpan->ui->type].create_io_watch(xetpan->ui, watcher)))
        {
          carray_delete(xetpan->ui->watched_fd, watcher_count - 1);
          free(watcher);
          return 0;
        }
    }

  return watcher->watch_id;
}

int
a_UI_remove_io_watch(XetPan *xetpan, unsigned int watch_id)
{
  XePUIIOWatcher *watcher;
  int n_watches;
  int ret = -1;
  int i;
  A_LOGGING_DEBUG("a_UI_remove_io_watch: removing watch_id %d", watch_id);
		 
  n_watches = carray_count(xetpan->ui->watched_fd);

  for (i = 0; i < n_watches; i++)
    {
      watcher = (XePUIIOWatcher *) carray_get(xetpan->ui->watched_fd, i);
      if (watcher->watch_id == watch_id)
	{
	  carray_delete(xetpan->ui->watched_fd, i);
          UI[xetpan->ui->type].delete_io_watch(xetpan->ui, watcher);
	  free(watcher);
	  ret = 0; /* Channel found */
	  break;
	}
    }

  if (ret == -1)
    A_LOGGING_SYSTEM_ERROR("no watch_id %d", watch_id);

  return ret;
}

/*! Returns the smallest unused watch_id.
 * As a collateral effect, it finds the maximal used 
 * watch_id value and stores it in ui->max_wid
 */
static unsigned int
UI_get_free_watch_id(XePUI *ui)
{
  unsigned int    max_wid;
  unsigned int    wid_size;
  unsigned int    watcher_count;
  unsigned int    wid_idx;
  char           *wid_map;
  unsigned int    i;
  XePUIIOWatcher *watcher;

  max_wid = ui->max_wid;
  
  wid_size = ((max_wid + 1) >> 3) + 1;

  wid_map = x_alloc(wid_size, char);

  watcher_count = carray_count(ui->watched_fd);

  for (i = 0; i < wid_size; i++)
    wid_map[0] = 0x00;

  /* watch_id = 0 is invalid */
  wid_map[0] = (char) 0x01;

  max_wid = 0;
  for (i = 0; i < watcher_count; i++)
    {
      watcher = (XePUIIOWatcher *) carray_get(ui->watched_fd, i);
      if (max_wid < watcher->watch_id)
	max_wid = watcher->watch_id;
      wid_idx = watcher->watch_id >> 3;
      wid_map[wid_idx] |= (char) (0x01 << (watcher->watch_id % 8));
    }
  ui->max_wid = max_wid;

  wid_idx = 0;

  /* Look for first not used watch_id.
   * We do not test for i because last byte is never 0xFF 
   */
  while (wid_map[wid_idx] == (char) 0xFF) 
    wid_idx++;

  i = 0;
  while (wid_map[wid_idx] & (0x01 << i))
    i++;

  free(wid_map);

  return (wid_idx << 3) + i;
}

