/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef UI_H
#define UI_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <libetpan/clist.h>

#include "ui-types.h"
#include "xetpan.h"

#ifdef USE_GTKUI
#include "gtk-ui.h"
#include "gtk-foldersview.h" 
#include "gtk-messageview.h" /* XePGTKMessageView */
#include "gtk-composewindow.h" /* XePGTKComposeWindow */
#endif
#ifdef USE_TTYUI
#include "tty-ui.h"
#endif

#include "message-types.h"
#include "mta.h"

/* Message View */

typedef struct XePUIMessageViewAny_s XePUIMessageViewAny;

struct XePUIMessageViewAny_s
{
  XePUIType type;      /**< UI type */
  XePUID uid;          /**< Unique Identifier for the message view */
  int editable;        /**< Wether editing is allowed */
  XetPan *xetpan;      /**< The global XetPan data structure */
  XePUIMain *uimain;   /**< The parent main window, if any */
  XePFolder *folder;   /**< The parent folder, if any */
  XePMessage *message; /**< Currently displayed message */
  uint32_t muid;       /**< muid of the currently displayed message, needed
			* because message are displayed in chunks,
			* asychronously (incremental reflow)
			*/
  clist *inlined_list; /**< List of XePUtilsInlinedData describing the
			*   inlined data
			*/
  char *body_text;  /**< Body text of composed message */
  char *default_charset; /**< Body's default charset */

};

union XePUIMessageView_u
{
  XePUIMessageViewAny any;
#ifdef USE_GTKUI
  XePGTKMessageView gtk_mview;
#endif
#ifdef USE_TTYUI
  XePTTYMessageView tty_mview;
#endif
};

/* Compose Window */

typedef struct XePUIComposeWindowAny_s XePUIComposeWindowAny;

struct XePUIComposeWindowAny_s
{
  XePUIType type;      /**< UI type                */
  XetPan *xetpan;      /**< The global XetPan data structure */
  XePUIMain *uimain;   /**< The parent main window if any */
  XePMTA *mta;         /**< The selected MTA */
  XePUIMessageView *messageview; 
  char *sent_folder;   /**< The folder that stores outgoing msg */
  /* The rest of the union is private to the UI */
};

union XePUIComposeWindow_u
{
  XePUIComposeWindowAny any;
#ifdef USE_GTKUI
  XePGTKComposeWindow gtk;
#endif
#ifdef USE_TTYUI
  XePTTYComposeWindow tty;
#endif
};

extern const char *A_UI_ABOUT_AUTHORS;
extern const char *A_UI_ABOUT_LICENSE;
extern const char *A_UI_ABOUT_HOME_PAGE;
extern const char *A_UI_ABOUT_COPYRIGHT;

/*! Create the main UI structure 
 *  This function created the XePUI main structure and initializes the
 *  static array UI that points to the specific UI main functions.
 *  \return the newly allocated XePUI structure ready to be used. The UI state 
 *          is set to 1 meaning that we are switching to a new UI (from no UI).
 */
XePUI *
a_UI_new(void);

/*! Frees the main UI structure. Frees only the resources it directly uses */
void
a_UI_free(XePUI *ui);

/*! Check if there is already an Id editor running
 *  \return 1 if an Id editor is already running, 0 otherwise
 */
int
a_UI_id_editor_is_locked(void);

/*! Set the Id editor flag, meaning an Id editor is being run
 *  \return 0 if lock succeeded, non 0 if the flag was alread locked
 */
int
a_UI_id_editor_lock(void);

/*! Unset the Id editor flag
 *  This function always succeeds
 */
void
a_UI_id_editor_unlock(void);

/*! UI Initialization :
 *  If UI type in XetPan::ui::type is specified, run only that UI and abort
 *  XetPan if an error occurs.
 *  If no UI type is specified, use the first registered UI 
 *  (currently GTK, TTY). If no UI could be initialized, display
 *  a message error.
 *  \param xetpan the mani XetPan structure that has its ui
 *         member defined.
 *  \return 0 on success and -1 if an error occured (no UI)
 */
int
a_UI_init(XetPan *xetpan, int *argc, char ***argv);

/*! UI main loop
 *  This function is called once and for all to run the UI main loop. 
 *  It returns only when XetPan is to be exited.
 *  It handles internally the UI switching (from TTY->GTK or from GTK->TTY)
 *  When a UI switch is requested, it calls initializes it and creates 
 *  a new main view, then it run the UI specific event loop.
 *  \param xetpan the main XetPan structure 
 */
void
a_UI_main(XetPan *xetpan);

/*! Ends the current UI event loop 
 *  This amounts to calling a_UI_main_switch() with the XETPAN_UI_QUIT
 *  UI type.
 *  \param xetpan the main XetPan structure
 */
void
a_UI_main_quit(XetPan *xetpan);

/*! Request a UI switch
 *  This function is to be called when the user switches from
 *  a UI event loop to another. Note that GTK event loop handles
 *  TTY events, but TTY event loop can't handle GTK events.
 *  When the same UI type as the current one is requested, nothing
 *  is done.
 *  The switch from TTY->GTK is needed when the user
 *  request a graphical interface. The switch from TTY->GTK
 *  is needed when the user will loose a graphical interface.
 *  (In the future, XetPan will be able to run in daemon mode)
 *  The switch is done by quitting the UI specifig event loop,
 *  and setting the UI switch flag. This function
 *  is also called by a_UI_main_quit() to quit XetPan, in
 *  which case it frees all IO handlers
 *  \param xetpan the main XetPan structure
 *  \param terminal the terminal/display where the new UI
 *         is to be opened
 *  \param uitype the type of the new UI (XETPAN_UI_TTY, XETPAN_UI_GTK
 *         or XETPAN_UI_QUIT to quit.
 *  \return 0 always...
 */
int
a_UI_main_switch(XetPan *xetpan, const char *terminal, int uitype);

/*! Check if there are some events that wait to be handled
 *  This function calls the specific UI function events_pending();
 *  \param xetpan the main XetPan structure
 *  \return 0 if no event is pending, 1 if some are waiting
 */
int
a_UI_events_pending(XetPan *xetpan);

/*! Run a single iteration of the event loop
 *  This function will wait for the first events and will process them
 *  by calling the appropriate associated callbacks
 *  \param xetpan the main XetPan structure 
 */
void
a_UI_main_iteration(XetPan *xetpan);

/*! Wait for certain semaphore to have a defined value, while running the event loop
 *  This function will not return until the semaphore has that value, without
 *  blocking the user interface. This can be helpful to sequentiallize some
 *  operations without blocking the UI
 *  \param xetpan the main XetPan structure
 *  \param semaphore a pointer to the variable which value is to be checked
 *  \param value the value to wait for
 */
void
a_UI_wait_until(XetPan *xetpan, int *semaphore, int value);

/*! Show the welcome window when the user launches XetPan for the first time
 *  This function is called when no configuration file is found. It
 *  greets the user and proposes to configure XetPan.
 *  After that, it should open a new main view or
 *  proceed with the composition of messages if the -c command
 *  switch was given.
 *  It calls the UI specific first_launch() function
 *  \param xetpan the main XetPan structure
 */
void
a_UI_first_launch(XetPan *xetpan);

void
a_UI_compose_pending_requests(XetPan *xetpan);

/*! Initialize and register a XePUIMain main view structure
 *  This function initializes an instance of the XePUIMain main view structure
 *  and registers it in the static global_mainviews table.
 */
void
a_UI_mainview_init(XePUIMain *uimain);


void
a_UI_foldersview_init(XePUIFoldersView *fview);

void
a_UI_foldersview_unregister(XePUIFoldersView *fview);

/*! Create and register a new XePUIComposeWindow structure
 *  This functions creates a new instance of a XePUIComposeWindow structure
 *  used to represent a compose window. Contrary to a_UI_mainview_new(),
 *  it does not register it. a_UI_composewindow_create() does it instead.
 *  This will be changed in the future.
 */
XePUIComposeWindow *
a_UI_composewindow_new(void);

/*! Create and register a new instance of XePUIMessageView
 *  This function creates a new instance of XePUIMessageView message
 *  view structure and registers is in the static global_messageviews
 *  table.
 */
XePUIMessageView *
a_UI_messageview_new(void);

/*! Create a new main view
 *  This function creates a new main view.
 *  A main view is the main user interface that provide access to all XetPan
 *  functionalities. It allows to select folder, view a summary of the 
 *  messages, view a message, run scripts, configure XetPan etc.
 *  \param xetpan the main XetPan structure
 *  \param terminal the terminal/display on which the main view is to
 *         be displayed. For X based UI, this corresponds to the
 *         display and has the same format as the $DISPLAY environment
 *         variable. $DISPLAY can be NULL, in which case the GUI 
 *         is responsible of using an appropriate display.
 *         For a TTY based UI, terminal consists of the
 *         terminal description in the same format as the $TERM
 *         environment variable along with the a tty path
 *         in the same format as returned by the ttyname() function,
 *         both elements being separated by a HTAB character (\t).
 *         If the terminal argument is NULL, the TTY interface
 *         will select the terminal bound to stdin and will get
 *         the terminal description from the $TERM environment variable
 *  \param type the UI type
 *  \return a newly allocated and registered XePUIMain. The main view
 *          has to be functional as soon as the UI event loop is 
 *          started.
 */
XePUIMain *
a_UI_mainview_create(XetPan *xetpan, char *terminal, int type);

/*! Close all UIs (windows) of a given UI type
 *  This function will close all windows/screens corresponding to a 
 *  certain UI type. The windows/screens are the main views
 *  and the compose windows. TODO: add also the standalone
 *  message views.
 *  \param xetpan the main XetPan structure
 *  \param uitype the UI type of which the windows are to be closed
 *  \return 1 if no more UI window/screen of any type is open, 0 otherwise
 */
int
a_UI_close_all(XetPan *xetpan, int uitype);

/*! Close a main view
 *  This function closes the given main view, by removing it from the
 *  list of available main views and by calling the UI specifig
 *  mainview_close() function.
 *  \param uimain the specific main view to close
 *  \return 1 if there are no more UI windows/screens, 0 other wise
 */
int
a_UI_mainview_close(XePUIMain *uimain);

/*! Close all main views of a given UI type
 *  This function will successively close all the main views
 *  of a certain UI type
 *  \param xetpan the main xetpan structure
 *  \param uitype the UI type of which the main views are to be closed
 *  \return if there are no more UI windows/screens, 0 other wise
 */
int
a_UI_mainview_close_all(XetPan *xetpan, int uitype);

/*! Notify all UI that foldersview list has been changed
 *  and reselect the summaryview
 */
void
a_UI_foldersview_list_updated(void);

/*! Refresh foldersview list data */
void
a_UI_foldersview_list_refresh(void);

/*! Expunge a folder then call a_UI_update_folder()
 *  This function requests the threads manager to update a folder. When thread
 *  finishes its work (the folder is updated), a_UI_folder_updated() is called
 *  If the request is done for an unknown folder, this functions issues an
 *  error message and returns without doing anything else.
 *  \param xetpan the main XetPan structure
 *  \param ref the ID of the folder to update
 */
void
a_UI_expunge_folder(const char *ref);

/*! Update a folder by rereading its content and refresh UIs that are displaying it.
 *  This function requests the threads manager to update a folder. When thread
 *  finishes its work (the folder is updated), a_UI_folder_updated() is called
 *  If the request is done for an unknown folder, this functions issues an
 *  error message and returns without doing anything else.
 *  \param xetpan the main XetPan structure
 *  \param ref the ID of the folder to update
 */
void
a_UI_update_folder(const char *ref);

/*! Update each UI that is displaying the given folder 
*   This functions checks each main view if it is
*   currently displaying the folder, in which case it 
*   will call the UI specific folder_updated() function
*   on that main view
*   \param axetpan the main XetPan structure
*   \param afolder the folder that was updated
*/
void
a_UI_folder_updated(void *anull, void *afolder);

void
a_UI_refresh_folder(const char *ref);

/*! Process a list of messages from a folder through a list of 
 *  processing rules.
 *  After processing, the folder is refreshed but not updated (FIXME).
 *  The foldersview lists are refreshed.
 *  \param folder the folder to which the messages pertain
 *  \param messages the array of messages to process. These must
 *         pertain to the aforementioned folder
 *  \param processing_rules a list of compiled processing nodes
 */
void
a_UI_folder_process(XePFolder *folder, 
		    carray *messages, 
		    clist *processing_rules);

/*! Update each UI that would be modified if the given message
 *  has changed state (flags)
 *  Function to call when flags of message have changed
 *  \param message the message that was modified
 */
void
a_UI_message_updated(XePMessage* message);

/*! Create a new compose window.
*   This function calls the UI specific composewindow_create()
*   and register the newly created instance of XePUIComposeWindow.
*   This function should not be called directly as it does not
*   take care of creating the underlying XePMessage structure.
*   Use one of the a_UI_composewindow_(compose|reply|forward|redirect)()
*   \param xetpan the main XetPan structure
*   \param terminal the terminal/display on which the window is
*          to be shown (same syntax as for a_UI_mainview_create())
*   \param type the UI type
*   \return a newly created instance of XePUIComposeWindow
*           compose window structure.
*/
XePUIComposeWindow *
a_UI_composewindow_create(XetPan *xetpan, char *terminal, int type);

/*! Create a new window to compose a new message
 *  This function creates a new window and prepares it for
 *  a new message to compose
 *  \param xetpan the main XetPan structure
 *  \param terminal the terminal/display where the window is
 *         to be opened
 *  \param uitype the UI type
 *  \param mailto a mailto URL to prefill differents parts of
 *         the new message
 *  \return a newly allocated and registered XePUIComposeWindow
 *          compose window structure
 */
XePUIComposeWindow *
a_UI_composewindow_compose(XetPan *xetpan, char *terminal, int type, 
			   const char *mailto);


/*! Create a new window to compose a reply
 *  This function creates a new window to reply to a
 *  an existant message
 *  \param xetpan the main XetPan structure
 *  \param terminal the terminal/display where the window is
 *         to be opened
 *  \param uitype the UI type
 *  \param original_message the original message to which the reply
 *         is created
 *  \param reply_text eventually the text to be quoted for the 
 *         reply.  The UI may decide to provide its own text
 *         derived or not from the original message text
 *  \param reply_charset the character set of the reply_text
 *  \return a newly allocated and registered XePUIComposeWindow
 *          compose window structure
 */
XePUIComposeWindow *
a_UI_composewindow_reply(XetPan *xetpan, 
			 char *terminal, 
			 int uitype,
			 XePMessage *message,
			 const char *reply_text,
			 const char *reply_charset);

XePUIComposeWindow *
a_UI_composewindow_forward(XetPan *xetpan, 
			   char *terminal, 
			   int uitype,
			   XePMessage *message);

XePUIComposeWindow *
a_UI_composewindow_redirect(XetPan *xetpan, 
			    char *terminal, 
			    int uitype,
			    XePMessage *message);
/*! Set the XePMessage to be edited.
 *  This function makes the specified compose window edit
 *  the given XePMessage
 */
void
a_UI_composewindow_set_message(XePUIComposeWindow *uicompose,
			       XePMessage *message);

void
a_UI_composewindow_set_mailto(XePUIComposeWindow *uicompose, 
			      const char *mailto);

void
a_UI_composewindow_set_body_text(XePUIComposeWindow *uicompose, 
				 const char *text, 
				 size_t text_size,
				 const char *text_charset);

void
a_UI_composewindow_add_header(XePUIComposeWindow *uicompose, 
			      XePMessageHeaderType header_type,
			      const char *header_name,
			      const char *header_value);

int
a_UI_composewindow_close(XePUIComposeWindow *uicompose);

int
a_UI_composewindow_close_all(XetPan *xetpan, int uitype);


/*! Create a standalone message viewer */
XePUIMessageView *
a_UI_messageview_create(XetPan *xetpan, 
			char *terminal, 
			int  uitype);

int
a_UI_messageview_close(XePUIMessageView *viewer);

int
a_UI_messageview_close_all(XetPan *xetpan, int uitype);

/*! View a RFC822 file in a message viewer */
void
a_UI_messageview_view_file(XePUIMessageView *mview, char *filename);


void
a_UI_notify_error(XetPan *xetpan);

void
a_UI_notify_error_reset(XetPan *xetpan);

/*! Submit job to change flags of an array of messages pertaining to a (REAL)
 *  folder
 *  \param folder the folder containing the messages
 *  \param messages an array of messages
 */
void
a_UI_folder_messages_change_flags(XePFolder *folder, carray *messages);

/*! Generic fuction to show a message in a message view. 
 */
void
a_UI_messageview_show_message(XePUIMessageView *mview,
			      XePMessage *message);

/*! Fetch a message and show it when available
 *  This function will request a thread to fetch a message,
 *  to find the default MIME part to show, to fetch that MIME part
 *  after which a first UI callback will show that MIME part, 
 *  a second one will asynchronously set the SEEN and unset the NEW flags
 *  of the message, and the third one 
 */
void
a_UI_messageview_fetch_show_async(XePUIMessageView *mview,
				  XePMessage *message);

void
a_UI_messageview_finalize(XePUIMessageView *mview, XePMessage *message);

/*! Change a message's flag and flush
 *  The XePMessage::flags_set and XePMessage::flags_unset must be set
 *  accordingly. This function will use a thread to change the flags
 *  Before setting them, all message views of the message are requested
 *  to block the mime view because their data will be invalidated by the
 *  mailmessage_flush. When the thread finishes its jobs, it will call
 *  a callback that will unblock the mime views.
 *  \param uimain the calling main window
 *  \param message the message for which the flags must be changed. The
 *         flags to set are to be stored in XePMessage::flags_set and the
 *         ones to unset are to be stored in XePMessage::flags_unset
 *  \return no value returned, this a an asynchroneous job
 */
void
a_UI_messageview_flags_change_async(XePUIMessageView *mview,
				    XePMessage *message);


/*! Request showing the MIME parts of a message
 *  Every UI that is currently showing this message will update its 
 *  MIME view.
 *  \param xetpan the main data pointer
 *  \param message the message which MIME view should be updated
 *  \return no value returned
 */
void
a_UI_messageview_mime_view_show(XetPan *xetpan, 
				XePMessage *message);


/*! Block the MIME view of a message
 *  Every UI that is currently showing this message will block its
 *  MIME view because the the MIME data are going to be invalidated
 *  \param xetpan the main data pointer
 *  \param message the message which MIME view should be blocked
 *  \return no value returned
 */
void
a_UI_messageview_mime_view_block(XetPan *xetpan,
				 XePMessage *message);


/*! Block the MIME view of a message
 *  Every UI that is currently showing this message will unblock its
 *  MIME view because the the MIME data are considered valid again.
 *  \param xetpan the main data pointer
 *  \param message the message which MIME view should be unblocked
 *  \return no value returned
 */
void
a_UI_messageview_mime_view_unblock(XetPan *xetpan, 
				   XePMessage *message);


/*! Fetch then show the selected MIME part
 *  A thread is requested to fetch the message part. When it is done
 *  the actual message viewing function is called
 *  \param the XePUIMessageView requesting the job
 *  \param the XePMessage that contains the selected MIME part to fetch 
 *         and show
 *  \return no value returned
 */
void
a_UI_messageview_mime_fetch_show_selected_async(XePUIMessageView *mview,
						XePMessage *message);


/*! Show the selected message part
 *  Only the requesting XePUIMain is concerned. 
 *  \param the XePUIMessageView that requested to show the MIME part
 *  \param message the message that contains the selected MIME part to show
 *  \return no value returned
 */
void
a_UI_messageview_mime_show_selected(XePUIMessageView *mview,
				    XePMessage *message);


/*! 
 */
void
a_UI_send_cmd_result(XetPan *xetpan);


/*! Add a handler when traffic occurs on a file descriptor 
 *  \param xetpan the main XetPan structure 
 *  \param fd the file descriptor to watch
 *  \param condition type of conditio to watch XETPAN_IO_WATCH_READ,
 *         XETPAN_IO_WATCH_WRITE or XETPAN_IO_WATCH_EXCEPTION
 *  \param callback a function that is fed with the file descriptor,
 *         the condition occured and the supplied user data and that
 *         must return non zero to stop watching
 *  \param data the user supplied data that will be passed to the callback
 *  \return the unique ID that can be used to stop watching.
 */
unsigned int
a_UI_add_io_watch(XetPan *xetpan,
		  int fd,
		  short int condition,
		  void (*callback)(int afd, short int acondition, void *adata),
		  void *data);

/*!
 */
int
a_UI_remove_io_watch(XetPan *xetpan, unsigned int watch_id);

#endif
